@@ -20,7 +20,9 @@ sed -i "s/^\(\(static \+\)\?\(const \+\)\?char .*_version..[ 	]*=[ 	]*\"\)[^\"]*
 sed -i "s/^\(\(static \+\)\?\(const \+\)\?char .*_date..[ 	]*=[ 	]*\"\)[^\"]*/\1${DATE}/" src/${PACKAGE}.c
 sed -i "/${YEAR}/!s/\(.* is copyright &copy; .*2[0-9]*\)\([ ]*Russell Stuart\)/\1,${YEAR}\2/" ${PACKAGE}.html
 sed -i "s/${PACKAGE}-[0-9]\+[.][0-9]\+[.][0-9]\+/${ME}/g" ${PACKAGE}.html
+sed -i "s/${ME%-*}-[0-9]\+[.][0-9]\+[.][0-9]\+/${ME}/g" ${PACKAGE}.html
 sed -i "/${YEAR}/!s/^\( *Copyright (c) .*2[0-9]*\)\([ ]*Russell Stuart\)/\1,${YEAR}\2/" README.txt
+sed -i "s/^\(\(version\|release\) *= *\).*/\1'${VERSION}'/" doc/conf.py
 
 #
 # Clean up.
@@ -35,7 +37,7 @@ rm -rf "${PACKAGE}"
 mkdir -p "${PACKAGE}"
 (cd ..; tar cfz "${ME}/${PACKAGE}/${ME}.tar.gz" --exclude="${ME}/debian" --exclude="${ME}/${PACKAGE}" "${ME}")
 cp ChangeLog.txt README.txt epl-v10.html "${PACKAGE}.html" "${PACKAGE}"
-cp -a doc/pam_python "${PACKAGE}/doc"
-cp -a doc/icons "${PACKAGE}/doc/icons"
+cp -a doc/html "${PACKAGE}/doc"
+cp -a examples "${PACKAGE}"
 ln -s "${PACKAGE}.html" "${PACKAGE}/index.html"
 cp "${PACKAGE}/${ME}.tar.gz" "../${ME%-*}_${VERSION}.orig.tar.gz"
