@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2007,2009 Russell Stuart
+ * Copyright (c) 2007,2009,2010 Russell Stuart
  *
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
@@ -35,8 +35,10 @@
 
 #define	PAMHANDLEEXCEPTION_NAME	"PamException"
 
-const char libpam_python_version[]	= "0.1.1";
-const char libpam_python_date[]		= "2009-08-05";
+#define arr_size(x)	(sizeof(x) / sizeof(*(x)))
+
+const char libpam_python_version[]	= "1.0.0";
+const char libpam_python_date[]		= "2010-06-18";
 
 /*
  * The python interpreter's shared library.
@@ -54,10 +56,10 @@ static void initialise_python(void)
   size_t		signum;
   struct sigaction	oldsigaction[NSIG];
 
-  for (signum = 0; signum < sizeof(oldsigaction)/sizeof(*oldsigaction); signum += 1)
+  for (signum = 0; signum < arr_size(oldsigaction); signum += 1)
     sigaction(signum, 0, &oldsigaction[signum]);
   Py_Initialize();
-  for (signum = 0; signum < sizeof(oldsigaction)/sizeof(*oldsigaction); signum += 1)
+  for (signum = 0; signum < arr_size(oldsigaction); signum += 1)
     sigaction(signum, &oldsigaction[signum], 0);
 #endif
 }
@@ -177,6 +179,7 @@ typedef struct
   int			py_initialized;	/* True if Py_initialize() called */
   PyTypeObject*		response;	/* pamh.Response */
   PyObject*		syslogFile;	/* A (the) SyslogFile instance */
+  PyTypeObject*		xauthdata;	/* pamh.XAuthData */
 } PamHandleObject;
 
 /*
@@ -215,7 +218,8 @@ static int SyslogFile_clear(PyObject* self)
 /*
  * Emulate python's file.write(), but write to syslog.
  */
-static PyObject* SyslogFile_write(PyObject* self, PyObject* args, PyObject* kwds)
+static PyObject* SyslogFile_write(
+    PyObject* self, PyObject* args, PyObject* kwds)
 {
   SyslogFileObject*	syslogFile = (SyslogFileObject*)self;
   const char*		c;
@@ -223,7 +227,7 @@ static PyObject* SyslogFile_write(PyObject* self, PyObject* args, PyObject* kwds
   int			len;
   const char*		newline;
   PyObject*		result = 0;
-  static char* kwlist[] = {"prompts", NULL};
+  static char* kwlist[] = {"data", NULL};
 
   if (!PyArg_ParseTupleAndKeywords(args, kwds, "s:write", kwlist, &data))
     goto error_exit;
@@ -364,7 +368,11 @@ static int syslog_path_exception(const char* module_path, const char* errormsg)
       str_message = PyString_AsString(message);
   }
   if (errormsg != 0 && str_name != 0 && str_message != 0)
-    syslog(LOG_AUTHPRIV|LOG_ERR, "%s - %s: %s", errormsg, str_name, str_message);
+  {
+    syslog(
+        LOG_AUTHPRIV|LOG_ERR, "%s - %s: %s",
+	errormsg, str_name, str_message);
+  }
   else if (str_name != 0 && str_message != 0)
     syslog(LOG_AUTHPRIV|LOG_ERR, "%s: %s", str_name, str_message);
   else if (errormsg != 0 && str_name != 0)
@@ -498,8 +506,7 @@ static int syslog_path_traceback(
  */
 static int syslog_traceback(PamHandleObject* pamHandle)
 {
-  return
-      syslog_path_traceback(get_module_path(pamHandle), pamHandle);
+  return syslog_path_traceback(get_module_path(pamHandle), pamHandle);
 }
 
 /*
@@ -544,13 +551,17 @@ static PyMemberDef PamMessage_members[] =
 static PyObject* PamMessage_new(
     PyTypeObject* type, PyObject* args, PyObject* kwds)
 {
+  int			err;
   PyObject*		msg = 0;
   int			msg_style = 0;
   PamMessageObject*	pamMessage = 0;
   PyObject*		self = 0;
   static char*		kwlist[] = {"msg_style", "msg", 0};
 
-  if (!PyArg_ParseTupleAndKeywords(args, kwds, "iO!:Message", kwlist, &msg_style, &PyString_Type, &msg))
+  err = PyArg_ParseTupleAndKeywords(
+      args, kwds, "iO!:Message", kwlist,
+      &msg_style, &PyString_Type, &msg);
+  if (!err)
     goto error_exit;
   pamMessage = (PamMessageObject*)type->tp_alloc(type, 0);
   if (pamMessage == 0)
@@ -607,13 +618,17 @@ static PyMemberDef PamResponse_members[] =
 static PyObject* PamResponse_new(
     PyTypeObject* type, PyObject* args, PyObject* kwds)
 {
+  int			err;
   PyObject*		resp = 0;
   int			resp_retcode = 0;
   PamResponseObject*	pamResponse = 0;
   PyObject*		self = 0;
   static char*		kwlist[] = {"resp", "resp_retcode", 0};
 
-  if (!PyArg_ParseTupleAndKeywords(args, kwds, "Oi:Response", kwlist, &resp, &resp_retcode))
+  err = PyArg_ParseTupleAndKeywords(
+      args, kwds, "Oi:Response", kwlist,
+      &resp, &resp_retcode);
+  if (!err)
     goto error_exit;
   if (resp != Py_None && !PyString_Check(resp))
   {
@@ -634,6 +649,73 @@ error_exit:
   return self;
 }
 
+/*
+ * The PamXAuthData object - used by PAM_XAUTHDATA item.
+ */
+#define	PAMXAUTHDATA_NAME	"XAuthData"
+typedef struct
+{
+  PyObject_HEAD				/* The Python Object header */
+  PyObject*		name;		/* struct pam_xauth_data.name */
+  PyObject*		data;		/* struct pam_xauth_data.data */
+} PamXAuthDataObject;
+
+static char PamXAuthData_doc[] =
+  MODULE_NAME "." PAMHANDLE_NAME "." PAMXAUTHDATA_NAME "(name, data)\n"
+  "  Constructs an immutable object is returned by and can be passed to\n"
+  "  the " MODULE_NAME ".xauthdata property.  The parameters are\n"
+  "  assigned to readonly members of the same name.";
+
+static PyMemberDef PamXAuthData_members[] =
+{
+  {
+    "data",
+    T_OBJECT_EX,
+    offsetof(PamXAuthDataObject, data),
+    READONLY,
+    "The value of the data item. A string or None.",
+  },
+  {
+    "name",
+    T_OBJECT_EX,
+    offsetof(PamXAuthDataObject, name),
+    READONLY,
+    "The name of the data item.  A string or None.",
+  },
+  {0,0,0,0,0},        	/* End of Python visible members */
+  {0,0,0,0,0}		/* Sentinal */
+};
+
+static PyObject* PamXAuthData_new(
+    PyTypeObject* type, PyObject* args, PyObject* kwds)
+{
+  int			err;
+  PyObject*		name = 0;
+  PyObject*		data = 0;
+  PamXAuthDataObject*	pamXAuthData = 0;
+  PyObject*		self = 0;
+  static char*		kwlist[] = {"name", "data", 0};
+
+  err = PyArg_ParseTupleAndKeywords(
+      args, kwds, "SS:XAuthData", kwlist,
+      &name, &data);
+  if (!err)
+    goto error_exit;
+  pamXAuthData = (PamXAuthDataObject*)type->tp_alloc(type, 0);
+  if (pamXAuthData == 0)
+    goto error_exit;
+  pamXAuthData->name = name;
+  Py_INCREF(pamXAuthData->name);
+  pamXAuthData->data = data;
+  Py_INCREF(pamXAuthData->data);
+  self = (PyObject*)pamXAuthData;
+  pamXAuthData = 0;
+
+error_exit:
+  py_xdecref((PyObject*)pamXAuthData);
+  return self;
+}
+
 /*
  * Check a PAM return value.  If the function failed raise an exception
  * and return -1.
@@ -688,12 +770,13 @@ error_exit:
 }
 
 static int PamHandle_set_item(
-    PyObject* self, int item_type, PyObject* pyValue)
+    PyObject* self, int item_type, char* item_name, PyObject* pyValue)
 {
   PamHandleObject*	pamHandle = (PamHandleObject*)self;
   int			pam_result;
   int			result = -1;
   char*			value;
+  char			error_message[64];
 
   if (pyValue == Py_None)
     value = 0;
@@ -702,7 +785,10 @@ static int PamHandle_set_item(
     value = PyString_AsString(pyValue);
     if (value == 0)
     {
-      PyErr_SetString(PyExc_TypeError, "PAM items must be set to strings");
+      snprintf(
+          error_message, sizeof(error_message),
+	  "PAM item %s must be set to a string", item_name);
+      PyErr_SetString(PyExc_TypeError, error_message);
       goto error_exit;
     }
     value = strdup(value);
@@ -910,7 +996,9 @@ static const char* PamEnv_getkey(PyObject* key)
   result = PyString_AS_STRING(key);
   if (*result == '\0')
   {
-    PyErr_SetString(PyExc_ValueError, "PAM environment key mustn't be 0 length");
+    PyErr_SetString(
+        PyExc_ValueError,
+	"PAM environment key mustn't be 0 length");
     return 0;
   }
   if (strchr(result, '=') != 0)
@@ -1071,6 +1159,7 @@ error_exit:
 static PyObject* PamEnv_get(
     PyObject* self, PyObject* args, PyObject* kwds)
 {
+  int			err;
   PamEnvObject*		pamEnv = (PamEnvObject*)self;
   PyObject*		default_value = 0;
   PyObject*		result = 0;
@@ -1079,7 +1168,10 @@ static PyObject* PamEnv_get(
   const char*		value_str;
   static char*		kwlist[] = {"key", "default", NULL};
 
-  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|O:get", kwlist, &key, &default_value))
+  err = PyArg_ParseTupleAndKeywords(
+      args, kwds, "O|O:get", kwlist,
+      &key, &default_value);
+  if (!err)
     goto error_exit;
   key_str = PamEnv_getkey(key);
   if (key_str == 0)
@@ -1268,6 +1360,12 @@ DECLARE_CONSTANT_GET(PAM_AUTHTOK_ERR)
 DECLARE_CONSTANT_GET(PAM_AUTHTOK_EXPIRED)
 DECLARE_CONSTANT_GET(PAM_AUTHTOK_LOCK_BUSY)
 DECLARE_CONSTANT_GET(PAM_AUTHTOK_RECOVER_ERR)
+#ifdef	PAM_AUTHTOK_RECOVERY_ERR
+DECLARE_CONSTANT_GET(PAM_AUTHTOK_RECOVERY_ERR)
+#endif
+#ifdef	PAM_AUTHTOK_TYPE
+DECLARE_CONSTANT_GET(PAM_AUTHTOK_TYPE)
+#endif
 DECLARE_CONSTANT_GET(PAM_BAD_ITEM)
 DECLARE_CONSTANT_GET(PAM_BINARY_PROMPT)
 DECLARE_CONSTANT_GET(PAM_BUF_ERR)
@@ -1321,6 +1419,12 @@ DECLARE_CONSTANT_GET(PAM_UPDATE_AUTHTOK)
 DECLARE_CONSTANT_GET(PAM_USER)
 DECLARE_CONSTANT_GET(PAM_USER_PROMPT)
 DECLARE_CONSTANT_GET(PAM_USER_UNKNOWN)
+#ifdef	PAM_XAUTHDATA
+DECLARE_CONSTANT_GET(PAM_XAUTHDATA)
+#endif
+#ifdef	PAM_XDISPLAY
+DECLARE_CONSTANT_GET(PAM_XDISPLAY)
+#endif
 
 #define	CONSTANT_GETSET(x) {#x,    PamHandle_Constant_ ## x, 0, 0, 0}
 
@@ -1333,10 +1437,13 @@ DECLARE_CONSTANT_GET(PAM_USER_UNKNOWN)
   static int PamHandle_set_##t(PyObject* self, PyObject* pyValue, void* closure) \
   { \
     closure = closure; \
-    return PamHandle_set_item(self, PAM_##t, pyValue); \
+    return PamHandle_set_item(self, PAM_##t, "PAM_" #t, pyValue); \
   }
 
 MAKE_GETSET_ITEM(AUTHTOK)
+#ifdef	PAM_AUTHTOK_TYPE
+MAKE_GETSET_ITEM(AUTHTOK_TYPE)
+#endif
 MAKE_GETSET_ITEM(OLDAUTHTOK)
 MAKE_GETSET_ITEM(RHOST)
 MAKE_GETSET_ITEM(RUSER)
@@ -1344,6 +1451,126 @@ MAKE_GETSET_ITEM(SERVICE)
 MAKE_GETSET_ITEM(TTY)
 MAKE_GETSET_ITEM(USER)
 MAKE_GETSET_ITEM(USER_PROMPT)
+#ifdef	PAM_XDISPLAY
+MAKE_GETSET_ITEM(XDISPLAY)
+#endif
+
+#ifdef	PAM_XAUTHDATA
+/*
+ * The PAM_XAUTHDATA item doesn't take strings like the rest of them.
+ * It wants a pam_xauth_data structure.
+ */
+static PyObject* PamHandle_get_XAUTHDATA(PyObject* self, void* closure)
+{
+  PamHandleObject*	pamHandle = (PamHandleObject*)self;
+  PyObject*		newargs = 0;
+  PyObject*		result = 0;
+  int			pam_result;
+  struct pam_xauth_data* xauth_data = 0;
+  
+  closure = closure;
+  pam_result = pam_get_item(
+      pamHandle->pamh, PAM_XAUTHDATA, (const void**)&xauth_data);
+  if (check_pam_result(pamHandle, pam_result) == -1)
+    goto error_exit;
+  if (xauth_data == 0)
+  {
+    result = Py_None;
+    Py_INCREF(result);
+  }
+  else
+  {
+    newargs = Py_BuildValue(
+        "s#s#",
+	xauth_data->name, xauth_data->namelen,
+	xauth_data->data, xauth_data->datalen);
+    if (newargs == 0)
+      goto error_exit;
+    result = pamHandle->xauthdata->tp_new(pamHandle->xauthdata, newargs, 0);
+    if (result == 0)
+      goto error_exit;
+  }
+
+error_exit:
+  py_xdecref(newargs);
+  return result;
+}
+
+static int PamHandle_set_XAUTHDATA(
+    PyObject* self, PyObject* pyValue, void* closure)
+{
+  PamHandleObject*	pamHandle = (PamHandleObject*)self;
+  PyObject*		name = 0;
+  PyObject*		data = 0;
+  int			result = -1;
+  const char*		data_str;
+  const char*		name_str;
+  int			pam_result;
+  struct pam_xauth_data	xauth_data;
+
+  closure = closure;
+  xauth_data.name = 0;
+  xauth_data.data = 0;
+  /*
+   * Get the name.
+   */
+  name = PyObject_GetAttrString(pyValue, "name");
+  if (name == 0)
+    goto error_exit;
+  name_str = PyString_AsString(name);
+  if (name_str == 0)
+  {
+    PyErr_SetString(PyExc_TypeError, "xauthdata.name must be a string");
+    goto error_exit;
+  }
+  xauth_data.name = strdup(name_str);
+  if (xauth_data.name == 0)
+  {
+    PyErr_NoMemory();
+    goto error_exit;
+  }
+  xauth_data.namelen = PyString_GET_SIZE(name);
+  /*
+   * Get the data.
+   */
+  data = PyObject_GetAttrString(pyValue, "data");
+  if (data == 0)
+    goto error_exit;
+  data_str = PyString_AsString(data);
+  if (data_str == 0)
+  {
+    PyErr_SetString(PyExc_TypeError, "xauthdata.data must be a string");
+    goto error_exit;
+  }
+  xauth_data.data = strdup(data_str);
+  if (xauth_data.data == 0)
+  {
+    PyErr_NoMemory();
+    goto error_exit;
+  }
+  xauth_data.datalen = PyString_GET_SIZE(data);
+  /*
+   * Set the item.  If that worked PAM will have swallowed the strings inside
+   * of it, so we must not free them.
+   */
+  pam_result = pam_set_item(pamHandle->pamh, PAM_XAUTHDATA, &xauth_data);
+  if (pam_result == PAM_SUCCESS)
+  {
+    xauth_data.name = 0;
+    xauth_data.data = 0;
+  }
+  result = check_pam_result(pamHandle, pam_result);
+
+error_exit:
+  py_xdecref(data);
+  py_xdecref(name);
+  if (xauth_data.name != 0)
+    free(xauth_data.name);
+  if (xauth_data.data != 0)
+    free(xauth_data.data);
+  return result;
+}
+#endif
 
 /*
  * Getters and setters.
@@ -1353,14 +1580,23 @@ static PyGetSetDef PamHandle_Getset[] =
   /*
    * Items.
    */
-  {"authtok",    PamHandle_get_AUTHTOK,    PamHandle_set_AUTHTOK,    "Authentication token", 0},
-  {"oldauthtok", PamHandle_get_OLDAUTHTOK, PamHandle_set_OLDAUTHTOK, "Old authentication token", 0},
-  {"rhost",      PamHandle_get_RHOST,      PamHandle_set_RHOST,      "Requesting host name", 0},
-  {"ruser",      PamHandle_get_RUSER,      PamHandle_set_RUSER,      "Requesting user name", 0},
-  {"service",    PamHandle_get_SERVICE,    PamHandle_set_SERVICE,    "Service (pam stack) name", 0},
-  {"tty",        PamHandle_get_TTY,        PamHandle_set_TTY,        "Terminal name", 0},
-  {"user",       PamHandle_get_USER,       PamHandle_set_USER,       "Authorized user name", 0},
-  {"user_prompt",PamHandle_get_USER_PROMPT,PamHandle_set_USER_PROMPT,"Prompt asking for users name", 0},
+  {"authtok",     PamHandle_get_AUTHTOK,     PamHandle_set_AUTHTOK,     "Authentication token", 0},
+#ifdef	PAM_AUTHTOK_TYPE
+  {"authtok_type",PamHandle_get_AUTHTOK_TYPE,PamHandle_set_AUTHTOK_TYPE,"XXX in the \"New XXX password:\" prompt", 0},
+#endif
+  {"oldauthtok",  PamHandle_get_OLDAUTHTOK,  PamHandle_set_OLDAUTHTOK,  "Old authentication token", 0},
+  {"rhost",       PamHandle_get_RHOST,       PamHandle_set_RHOST,       "Requesting host name", 0},
+  {"ruser",       PamHandle_get_RUSER,       PamHandle_set_RUSER,       "Requesting user name", 0},
+  {"service",     PamHandle_get_SERVICE,     PamHandle_set_SERVICE,     "Service (pam stack) name", 0},
+  {"tty",         PamHandle_get_TTY,         PamHandle_set_TTY,         "Terminal name", 0},
+  {"user",        PamHandle_get_USER,        PamHandle_set_USER,        "Authorized user name", 0},
+  {"user_prompt", PamHandle_get_USER_PROMPT, PamHandle_set_USER_PROMPT, "Prompt asking for users name", 0},
+#ifdef	PAM_XAUTHDATA
+  {"xauthdata",	  PamHandle_get_XAUTHDATA,   PamHandle_set_XAUTHDATA,   "The name of the X display ($DISPLAY)", 0},
+#endif
+#ifdef	PAM_XDISPLAY
+  {"xdisplay",	  PamHandle_get_XDISPLAY,    PamHandle_set_XDISPLAY,    "The name of the X display ($DISPLAY)", 0},
+#endif
   /*
    * Constants.
    */
@@ -1375,6 +1611,12 @@ static PyGetSetDef PamHandle_Getset[] =
   CONSTANT_GETSET(PAM_AUTHTOK_EXPIRED),
   CONSTANT_GETSET(PAM_AUTHTOK_LOCK_BUSY),
   CONSTANT_GETSET(PAM_AUTHTOK_RECOVER_ERR),
+#ifdef	PAM_AUTHTOK_RECOVERY_ERR
+  CONSTANT_GETSET(PAM_AUTHTOK_RECOVERY_ERR),
+#endif
+#ifdef	PAM_AUTHTOK_TYPE
+  CONSTANT_GETSET(PAM_AUTHTOK_TYPE),
+#endif
   CONSTANT_GETSET(PAM_BAD_ITEM),
   CONSTANT_GETSET(PAM_BINARY_PROMPT),
   CONSTANT_GETSET(PAM_BUF_ERR),
@@ -1428,6 +1670,10 @@ static PyGetSetDef PamHandle_Getset[] =
   CONSTANT_GETSET(PAM_USER),
   CONSTANT_GETSET(PAM_USER_PROMPT),
   CONSTANT_GETSET(PAM_USER_UNKNOWN),
+  CONSTANT_GETSET(PAM_XAUTHDATA),
+#ifdef	PAM_XDISPLAY
+  CONSTANT_GETSET(PAM_XDISPLAY),
+#endif
   {0,0,0,0,0}        	/* Sentinel */
 };
 
@@ -1494,6 +1740,7 @@ error_exit:
 static PyObject* PamHandle_conversation(
     PyObject* self, PyObject* args, PyObject* kwds)
 {
+  int			err;
   PamHandleObject*	pamHandle = (PamHandleObject*)self;
   PyObject*		prompts = 0;
   PyObject*		result_tuple = 0;
@@ -1510,7 +1757,10 @@ static PyObject* PamHandle_conversation(
   int			py_result;
   static char*		kwlist[] = {"prompts", NULL};
 
-  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O:conversation", kwlist, &prompts))
+  err = PyArg_ParseTupleAndKeywords(
+      args, kwds, "O:conversation", kwlist,
+      &prompts);
+  if (!err)
     goto error_exit;
   pam_result = pam_get_item(pamHandle->pamh, PAM_CONV, (const void**)&conv);
   if (check_pam_result(pamHandle, pam_result) == -1)
@@ -1575,7 +1825,8 @@ static PyObject* PamHandle_conversation(
       goto error_exit;
     for (i = 0; i < prompt_count; i += 1)
     {
-      response = PamHandle_conversation_2response(pamHandle, &response_array[i]);
+      response = PamHandle_conversation_2response(
+          pamHandle, &response_array[i]);
       if (response == 0)
         goto error_exit;
       if (PyTuple_SetItem(result_tuple, i, response) == -1)
@@ -1602,13 +1853,17 @@ error_exit:
 static PyObject* PamHandle_fail_delay(
     PyObject* self, PyObject* args, PyObject* kwds)
 {
+  int			err;
   PamHandleObject*	pamHandle = (PamHandleObject*)self;
   int			micro_sec = 0;
   int			pam_result;
   PyObject*		result = 0;
   static char*		kwlist[] = {"micro_sec", NULL};
 
-  if (!PyArg_ParseTupleAndKeywords(args, kwds, "i:fail_delay", kwlist, &micro_sec))
+  err = PyArg_ParseTupleAndKeywords(
+      args, kwds, "i:fail_delay", kwlist,
+      &micro_sec);
+  if (!err)
     goto error_exit;
   pam_result = pam_fail_delay(pamHandle->pamh, micro_sec);
   if (check_pam_result(pamHandle, pam_result) == -1)
@@ -1655,7 +1910,8 @@ error_exit:
 /*
  * Set a PAM environment variable.
  */
-static PyObject* PamHandle_strerror(PyObject* self, PyObject* args, PyObject* kwds)
+static PyObject* PamHandle_strerror(
+    PyObject* self, PyObject* args, PyObject* kwds)
 {
   PamHandleObject*	pamHandle = (PamHandleObject*)self;
   const char*		err;
@@ -1789,6 +2045,13 @@ static PyMemberDef PamHandle_Members[] =
     READONLY,
     "Response class returned by " MODULE_NAME "." PAMHANDLE_NAME ".conversation()"
   },
+  {
+    "XAuthData",
+    T_OBJECT,
+    offsetof(PamHandleObject, xauthdata),
+    READONLY,
+    "XAuthData class used by " MODULE_NAME "." PAMHANDLE_NAME ".xauthdata"
+  },
   {0,0,0,0,0},        	/* End of Python visible members */
   {
     "syslogFile",
@@ -1864,7 +2127,9 @@ static int load_user_module(
   module_fp = fopen(module_path, "r");
   if (module_fp == 0)
   {
-    syslog_path_message(module_path, "Can not open module: %s", strerror(errno));
+    syslog_path_message(
+        module_path, "Can not open module: %s",
+	strerror(errno));
     pam_result = PAM_OPEN_ERR;
     goto error_exit;
   }
@@ -2036,7 +2301,8 @@ static PyObject* newSingletonObject(
 
 /*
  * Find the PamHandle object used by the pamh instance, creating one if it
- * doesn't exist.
+ * doesn't exist.  Returns a pam_result, which will be PAM_SUCCESS if it
+ * works.
  */
 static int get_pamHandle(
   PamHandleObject** result, pam_handle_t* pamh, const char** argv)
@@ -2184,7 +2450,12 @@ static int get_pamHandle(
   if (pamEnv->pamEnvIter_type == 0)
     goto error_exit;
   if (PyObject_IS_GC((PyObject*)pamEnv->pamEnvIter_type))
-    PyObject_GC_UnTrack(pamEnv->pamEnvIter_type); /* No refs are visible to python */
+  {
+    /*
+     * No refs are visible to python.
+     */
+    PyObject_GC_UnTrack(pamEnv->pamEnvIter_type);
+  }
   pamEnv->pamEnvIter_type->tp_iter = PyObject_SelfIter;
   pamEnv->pamEnvIter_type->tp_iternext = PamEnvIter_iternext;
   pamHandle->env = (PyObject*)pamEnv;
@@ -2204,7 +2475,8 @@ static int get_pamHandle(
       PamMessage_new);			/* tp_new */
   if (pamHandle->message == 0)
   {
-    pam_result = syslog_path_exception(module_path, "Can't create pamh.Message");
+    pam_result = syslog_path_exception(
+        module_path, "Can't create pamh.Message");
     goto error_exit;
   }
   /*
@@ -2271,6 +2543,25 @@ static int get_pamHandle(
     goto error_exit;
   }
   Py_INCREF(pamHandle->print_exception); /* Borrowed reference */
+  /*
+   * Create the type for the PamXAuthDataObject.
+   */
+  pamHandle->xauthdata = newHeapType(
+      pamHandle_module,			/* __module__ */
+      PAMXAUTHDATA_NAME "_type",	/* tp_name */
+      sizeof(PamXAuthDataObject),	/* tp_basicsize */
+      PamXAuthData_doc,			/* tp_doc */
+      0,				/* tp_clear */
+      0,				/* tp_methods */
+      PamXAuthData_members,		/* tp_members */
+      0,				/* tp_getset */
+      PamXAuthData_new);		/* tp_new */
+  if (pamHandle->xauthdata == 0)
+  {
+    pam_result = syslog_path_exception(
+        module_path, "Can't create pamh.XAuthData");
+    goto error_exit;
+  }
   /*
    * Now we have error reporting set up import the module.
    */
@@ -2404,7 +2695,7 @@ static int call_handler(
   int			pam_result;
 
   /*
-   * Initialise Python, and get a copy of out object.
+   * Initialise Python, and get a copy of our object.
    */
   pam_result = get_pamHandle(&pamHandle, pamh, argv);
   if (pam_result != PAM_SUCCESS)
