@@ -3,6 +3,11 @@
 # This is the test script for libpython-pam.  There aren't many stones
 # left unturned.
 #
+# Best run from the Makefile using the target 'test'.  To run manually:
+#   sudo ln -s $PWD/test-pam_python.pam /etc/pam.d
+#   python test.py
+#   sudo rm /etc/pam.d/test-pam_python.pam 
+#
 import os
 import sys
 
@@ -38,6 +43,7 @@ def run_test(caller):
   import test
   test_name = caller.__name__[4:]
   sys.stdout.write("Testing " + test_name + " ")
+  sys.stdout.flush()
   test.test_results = []
   test.test_function = globals()["test_" + test_name]
   caller(test.test_results)
@@ -46,6 +52,17 @@ def run_test(caller):
 def pam_conv(auth, query_list, userData):
   return query_list
 
+#
+# Verify the results match.
+#
+def assert_results(expected_results, results):
+  for i in range(min(len(expected_results), len(results))):
+    assert expected_results[i] == results[i], (i, expected_results[i], results[i])
+  if len(expected_results) < len(results):
+    assert len(expected_results) == len(results), (i, results[len(expected_results)])
+  else:
+    assert len(expected_results) == len(results), (i, expected_results[len(results)])
+
 #
 # Test all the calls happen.
 #
@@ -71,8 +88,7 @@ def run_basic_calls(results):
       (pam_sm_open_session.func_name, 0, [me]),
       (pam_sm_close_session.func_name, 0, [me]),
       (pam_sm_end.func_name, None, None)]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Test all the constants are defined.
@@ -103,6 +119,7 @@ PAM_CONSTANTS = {
     "PAM_CONV_ERR":			19,
     "PAM_AUTHTOK_ERR":			20,
     "PAM_AUTHTOK_RECOVER_ERR":		21,
+    "PAM_AUTHTOK_RECOVERY_ERR":		21,
     "PAM_AUTHTOK_LOCK_BUSY":		22,
     "PAM_AUTHTOK_DISABLE_AGING":	23,
     "PAM_TRY_AGAIN":			24,
@@ -123,6 +140,9 @@ PAM_CONSTANTS = {
     "PAM_RUSER":			8,
     "PAM_USER_PROMPT":			9,
     "PAM_FAIL_DELAY":			10,
+    "PAM_XDISPLAY":			11,
+    "PAM_XAUTHDATA":			12,
+    "PAM_AUTHTOK_TYPE":			13,
     "PAM_SILENT":			0x8000,
     "PAM_DISALLOW_NULL_AUTHTOK":	0x0001,
     "PAM_ESTABLISH_CRED":		0x0002,
@@ -157,6 +177,11 @@ def test_constants(results, who, pamh, flags, argv):
       for var in dir(pamh)
       if var.startswith("PAM_") or var.startswith("_PAM_")])
   results.append(pam_constants)
+  try:
+    pamh.PAM_SUCCESS = 1
+    results.append("Opps, pamh.PAM_SUCCESS = 1 worked!")
+  except StandardError, e:
+    results.append("except: %s" % e)
   return pamh.PAM_SUCCESS
 
 def run_constants(results):
@@ -165,21 +190,18 @@ def run_constants(results):
   pam.authenticate(0)
   pam.close_session()
   del pam
-  assert results[0] == pam_sm_authenticate.func_name
-  assert results[2] == pam_sm_close_session.func_name
-  assert results[3] == pam_sm_end.func_name
+  assert results[0] == pam_sm_authenticate.func_name, (results[0], pam_sm_authenticate.func_name)
+  assert results[2] == "except: attribute 'PAM_SUCCESS' of 'PamHandle_type' objects is not writable", results[2]
+  assert results[3] == pam_sm_close_session.func_name, (results[3], pam_sm_close_session.func_name)
+  assert results[4] == pam_sm_end.func_name, (results[4], pam_sm_end.func_name)
   consts = results[1]
   for var in PAM_CONSTANTS.keys():
-    if not consts.has_key(var): print var
-    assert consts.has_key(var)
-    if not consts[var] == PAM_CONSTANTS[var]: print var
-    assert consts[var] == PAM_CONSTANTS[var]
+    assert consts.has_key(var), var
+    assert consts[var] == PAM_CONSTANTS[var], (var, consts[var], PAM_CONSTANTS[var])
   for var in consts.keys():
-    if not PAM_CONSTANTS.has_key(var): print var
-    assert PAM_CONSTANTS.has_key(var)
-    if not PAM_CONSTANTS[var] == consts[var]: print var
-    assert PAM_CONSTANTS[var] == consts[var]
-  assert len(results) == 4
+    assert PAM_CONSTANTS.has_key(var), var
+    assert PAM_CONSTANTS[var] == consts[var], (var, PAM_CONSTANTS[var], consts[var])
+  assert len(results) == 5, len(results)
 
 #
 # Test the environment calls.
@@ -254,8 +276,7 @@ def run_environment(results):
       ['x2', 'x3', 'xx'],
       ['2', '3', 'x'],
       pam_sm_close_session.func_name, pam_sm_end.func_name]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Test strerror().
@@ -275,12 +296,11 @@ def run_strerror(results):
   expected_results = [
       pam_sm_authenticate.func_name,
       ( 0, 'success'),
-      ( 1, 'dlopen() failure'),
+      ( 1, 'failed to load module'),
       (30, 'conversation is waiting for event'),
       (31, 'application needs to call libpam again'),
       pam_sm_end.func_name]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Test items.
@@ -291,12 +311,14 @@ def test_items(results, who, pamh, flags, argv):
     return pamh.PAM_SUCCESS
   items = {
 	"authtok":	"authtok-module",
+	"authtok_type":	"authtok_type-module",
 	"oldauthtok":	"oldauthtok-module",
 	"rhost":	"rhost-module",
 	"ruser":	"ruser-module",
 	"tty":		"tty-module",
-	"user":		"user-module",
 	"user_prompt":	"user_prompt-module",
+	"user":		"user-module",
+	"xdisplay":	"xdisplay-module",
       }
   keys = items.keys()
   keys.sort()
@@ -305,6 +327,11 @@ def test_items(results, who, pamh, flags, argv):
     value = items[key]
     if value != None:
       setattr(pamh, key, value)
+  try:
+    setattr(pamh, "tty", 1)
+    results.append("%r = %r" % (key, value))
+  except StandardError, e:
+    results.append("except: %s" % e)
   results.append(pamh.get_user("a prompt"))
   return pamh.PAM_SUCCESS
 
@@ -317,27 +344,105 @@ def run_items(results):
       3:	"tty",
       4:	"rhost",
       8:	"ruser",
-      9:	"user_prompt"}
-  for item, value in items.items():
-    pam.set_item(item, value)
+      9:	"user_prompt",
+      11:	"xdisplay",
+      13:	"authtok_type"}
+  items_list = items.keys()
+  items_list.sort()
+  for item in items_list:
+    pam.set_item(item, items[item])
   pam.open_session()
   pam.close_session()
   del pam
   expected_results = [
       pam_sm_authenticate.func_name, pam_sm_open_session.func_name,
-      ('authtok', None), ('oldauthtok', None),
-      ('rhost','rhost'), ('ruser', 'ruser'), ('tty', 'tty'),
-      ('user', 'user'), ('user_prompt', 'user_prompt'),
+      ('authtok',	None),
+      ('authtok_type',	'authtok_type'),
+      ('oldauthtok',	None),
+      ('rhost',		'rhost'),
+      ('ruser',		'ruser'),
+      ('tty',		'tty'),
+      ('user',		'user'),
+      ('user_prompt',	'user_prompt'),
+      ('xdisplay',	'xdisplay'),
+      'except: PAM item PAM_TTY must be set to a string',
       'user-module',
       pam_sm_close_session.func_name,
-      ('authtok', 'authtok-module'), ('oldauthtok', 'oldauthtok-module'),
-      ('rhost', 'rhost-module'), ('ruser', 'ruser-module'),
-      ('tty', 'tty-module'), ('user', 'user-module'),
-      ('user_prompt', 'user_prompt-module'),
+      ('authtok',	'authtok-module'),
+      ('authtok_type',	'authtok_type-module'),
+      ('oldauthtok',	'oldauthtok-module'),
+      ('rhost',		'rhost-module'),
+      ('ruser',		'ruser-module'),
+      ('tty',		'tty-module'),
+      ('user',		'user-module'),
+      ('user_prompt',	'user_prompt-module'),
+      ('xdisplay',	'xdisplay-module'),
+      'except: PAM item PAM_TTY must be set to a string',
       'user-module',
       pam_sm_end.func_name]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
+
+#
+# Test the xauthdata item.
+#
+def test_xauthdata(results, who, pamh, flags, argv):
+  results.append(who.func_name)
+  if not who in (pam_sm_open_session, pam_sm_close_session):
+    return pamh.PAM_SUCCESS
+  xauthdata0 = pamh.XAuthData("name-module", "data-module")
+  pamh.xauthdata = xauthdata0
+  xauthdata1 = pamh.xauthdata
+  results.append('name=%r, data=%r' % (xauthdata1.name, xauthdata1.data))
+  try:
+    xauthdata2 = pamh.XAuthData(None, "x")
+    results.append('pamh.XAuthData(%r, %r)' % (xauthdata2.name, xauthdata2.data))
+  except TypeError, e:
+    results.append('except: %s' % e)
+  try:
+    xauthdata2 = pamh.XAuthData("x", 1)
+    results.append('pamh.XAuthData(%r, %r)' % (xauthdata2.name, xauthdata2.data))
+  except TypeError, e:
+    results.append('except: %s' % e)
+  class XA: pass
+  XA.name = "name-XA"
+  XA.data = "data-XA"
+  pamh.xauthdata = XA
+  xauthdata2 = pamh.xauthdata
+  results.append('name=%r, data=%r' % (xauthdata2.name, xauthdata2.data))
+  xa = XA()
+  xa.name = "name-xa"
+  xa.data = "data-xa"
+  pamh.xauthdata = xa
+  xauthdata4 = pamh.xauthdata
+  results.append('name=%r, data=%r' % (xauthdata4.name, xauthdata4.data))
+  return pamh.PAM_SUCCESS
+
+def run_xauthdata(results):
+  pam = PAM.pam()
+  pam.start(TEST_PAM_MODULE, TEST_PAM_USER, pam_conv)
+  pam.authenticate(0)
+  #
+  # The PAM module doesn't support XAUTHDATA, so check what we can from the
+  # module only.
+  #
+  pam.open_session()
+  pam.close_session()
+  del pam
+  expected_results = [
+      pam_sm_authenticate.func_name, pam_sm_open_session.func_name,
+      ("name='name-module', data='data-module'"),
+      'except: XAuthData() argument 1 must be string, not None',
+      'except: XAuthData() argument 2 must be string, not int',
+      ("name='name-XA', data='data-XA'"),
+      ("name='name-xa', data='data-xa'"),
+      pam_sm_close_session.func_name,
+      ("name='name-module', data='data-module'"),
+      'except: XAuthData() argument 1 must be string, not None',
+      'except: XAuthData() argument 2 must be string, not int',
+      ("name='name-XA', data='data-XA'"),
+      ("name='name-xa', data='data-xa'"),
+      pam_sm_end.func_name]
+  assert_results(expected_results, results)
 
 #
 # Test having no pam_sm_end.
@@ -354,8 +459,7 @@ def run_no_sm_end(results):
   pam.authenticate(0)
   del pam
   expected_results = [pam_sm_authenticate.func_name]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Test the conversation mechanism.
@@ -401,8 +505,7 @@ def run_conv(results):
       pam_sm_acct_mgmt.func_name,
       ('single', 1),
       pam_sm_end.func_name]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Test pam error returns.
@@ -422,8 +525,7 @@ def run_pamerr(results):
   del pam
   expected_results = [-r for r in range(PAM._PAM_RETURN_VALUES)]
   expected_results[25] = -6
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Test fail_delay.
@@ -467,8 +569,7 @@ def run_exceptions(results):
   del pam
   expected_results = [results[0], 0]
   expected_results += [(-r,) for r in range(1, results[0])]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Test absent entry point.
@@ -505,14 +606,13 @@ def run_absent(results):
   del pam
   expected_results = [
       'pam_sm_authenticate',
-      ('error', 'Not authenticated'),	# PyPAM doesn't return the real error
-      ('error', 'Not authenticated'),
-      ('error', 'Not authenticated'),
-      ('error', 'Not authenticated'),
-      ('error', 'Not authenticated')
+      ('error', "('Symbol not found', 2)"),
+      ('error', "('Symbol not found', 2)"),
+      ('error', "('Symbol not found', 2)"),
+      ('error', "('Symbol not found', 2)"),
+      ('error', "('Symbol not found', 2)"),
     ]
-  if not expected_results == results: print results; print expected_results
-  assert expected_results == results
+  assert_results(expected_results, results)
 
 #
 # Entry point.
@@ -523,6 +623,7 @@ def main(argv):
   run_test(run_environment)
   run_test(run_strerror)
   run_test(run_items)
+  run_test(run_xauthdata)
   run_test(run_no_sm_end)
   run_test(run_conv)
   run_test(run_pamerr)
